#
# SSM_Tools::Database -- SSM_Tools Database module 
#
# Copyright (C) 2017-2021 Francesco P. Lovergine <francesco.lovergine@cnr.it>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2, 
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
#

#
# See POD at the __END__ for documentation
#
package SSM_Tools::Database;

use v5.18;

use strict;

use warnings;
no warnings 'experimental::smartmatch';

use DBI;
use Carp;
use File::Basename;
use File::Copy;
use File::Path qw/make_path/;

use Data::Dumper;
$Data::Dumper::Sortkeys = 1;

require Exporter;
use vars qw($VERSION @ISA @EXPORT @EXPORT_OK %EXPORT_TAGS);

# set the version for version checking
$VERSION     = '0.02';

@ISA         = qw(Exporter);
@EXPORT      = qw();

%EXPORT_TAGS = ( );     # eg: TAG => [ qw!name1 name2! ],

# your exported package globals go here,
# as well as any optionally exported functions
@EXPORT_OK   = qw($Verbose $Replace $Link);

use vars qw($Verbose $Replace $Link);

# non-exported package globals go here
use vars qw();

# initialize package globals, first exported ones

$Verbose = 0;
$Replace = 0;
$Link = 1;

my $_replace = 'IGNORE';
$_replace = 'REPLACE' if $Replace;

# then the others (which are still accessible as $Some::Module::stuff)

# all file-scoped lexicals must be created before
# the functions below that use them.

# file-private lexicals go here

use constant DBNAME => qq(/usr/local/sm-tools/data/es1.splite);

sub new($$) {
	my ($class, $dbname) = @_;
	$dbname //= DBNAME;
	
	my $self = {};
	my $dbh = DBI->connect(
		"DBI:SQLite:dbname=$dbname",
		'',
		'',
		{	
				PrintError => 0,
				RaiseError => 1,
				AutoCommit => 0,
		}
	) or Carp::croak $DBI::errstr;
	
	# load Spatialite extension
  	$dbh->sqlite_enable_load_extension(1);
	# this extension load instruction is platform specific...
	my $sth = $dbh->prepare(qq/SELECT load_extension('libspatialite.so')/);
	eval { $sth->execute; };
	if ($@) {
		# in version 4.3 extension changed name
		$sth = $dbh->prepare(qq/SELECT load_extension('mod_spatialite')/);
		eval { $sth->execute; };
		if ($@) {
			$sth = $dbh->prepare(qq/SELECT load_extension('mod_spatialite.so')/);
			eval { $sth->execute; };
			Carp::croak qq/Missing Spatialite extension or wrong extensiona name/ if $@;
		}
	}
	
	# check if Spatialite metatables exist and eventually force their creation
	$sth = $dbh->prepare(qq/SELECT count(*) FROM sqlite_master WHERE 
						type='table' AND name='spatial_ref_sys'/);
	$sth->execute;
	my $row = $sth->fetchrow_arrayref;
	if (!$row || !$row->[0]) {
		$sth = $dbh->prepare(qq/SELECT InitSpatialMetaData()/);
		$sth->execute;
		$dbh->commit;
	}
	# The default cache is quite low
	$dbh->do(qq/PRAGMA cache_size = 100000/);
	$dbh->do(qq/PRAGMA page_size = 4096/);
	$dbh->do(qq/PRAGMA locking_mode=EXCLUSIVE/);
	$dbh->do(qq/PRAGMA temp_store=MEMORY/);
	$self->{dbh} = $dbh;
	$self->{timestamp} = undef;
	bless $self, $class;
	$_replace = 'REPLACE' if $Replace;
	return $self;
}

sub _trim($$) {
	my $str = $_[1] =~ s/^\s+|\s+$//gr;
	$str;
}

sub _timestamp($$) {
	my ($self, $time) = @_;

	$time ||= time;
    my ($sec,$min ,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = gmtime($time);
    sprintf("%04d-%02d-%02d %02d:%02d:%02d",$year+1900,$mon+1,$mday,$hour,$min,$sec);
}

sub _depth($$) {
	my ($self, $str) = @_;
	$str //= "";
	$str =~ /\s*(\d+)?(?:[dD](\d+))?(?:[tT](\d+)(?:[dD](\d+))?)?(?::(\d+))?\s*/;
	my @depth = ( ($1 ? "$1" : "") . ($2 ? ".$2" : ""), 
				  ($3 ? "$3" : "") . ($4 ? ".$4" : ""),
				  ($5 ? "$5" : "")
			    );
	@depth;
}

sub _sensor_names($$) {
	my ($self, $str) = @_;
	my @alias = split(/:/, $str);
	@alias;
}

sub _site_check_in($$) {
	my ($self, $site) = @_;
	my $dbh = $self->{dbh};
	my $id = undef;

	my $sth = $dbh->prepare(qq/SELECT id FROM Site WHERE name=?/);
	$sth->execute(lc($site));
	my $row = $sth->fetchrow_arrayref;
	$sth->finish;
	if (!$row) {
		$sth = $dbh->prepare(qq/INSERT INTO Site (id,name) VALUES (NULL,?)/);
		$sth->execute(lc($site));
		$id = $dbh->last_insert_id(undef, undef, undef, undef);
		warn qq/Site $site not found, inserted a new one $id/ if $Verbose;
		$dbh->commit;
	} else {
		$id = $row->[0];
	}
	$id;
}

sub _path_check_in($$) {
	my ($self, $uri) = @_;
	my $dbh = $self->{dbh};
	my $id = undef;

	my $sth = $dbh->prepare(qq/SELECT id FROM Path WHERE uri=?/);
	$sth->execute($uri);
	my $row = $sth->fetchrow_arrayref;
	$sth->finish;
	if (!$row) {
		$sth = $dbh->prepare(qq/INSERT INTO Path (id,uri) VALUES (NULL,?)/);
		$sth->execute($uri);
		$id = $dbh->last_insert_id(undef, undef, undef, undef);
		warn qq/Path $uri not found, inserted a new one as $id/ if $Verbose;
		$dbh->commit;
	} else {
		$id = $row->[0];
	}
	$id;
}

sub _network_check_in($$$) {
	my ($self, $site_id, $network) = @_;
	my $dbh = $self->{dbh};
	my $id = undef;

	my $sth = $dbh->prepare(qq/
		SELECT n.id, s.id FROM 
			Network AS n, Site AS s 
		WHERE s.id=? AND n.name=? AND s.id = n.site_id
	/);
	$sth->execute($site_id, lc($network));
	my $row = $sth->fetchrow_arrayref;
	$sth->finish;
	if (!$row) {
		$sth = $dbh->prepare(qq/INSERT INTO Network (id,name,site_id) VALUES (NULL,?,?)/);
		$sth->execute(lc($network), $site_id);
		$id = $dbh->last_insert_id(undef, undef, undef, undef);
		warn qq/Network $network not found, inserted a new one $id/ if $Verbose;
		$dbh->commit;
	} else {
		$id = $row->[0];
	}
	$id;
}

sub _sensor_check_in($$$$$$$$) {
	my ($self, $site_id, $network_id, $sensor, $long, $lat, $a, $b) = @_;
	my $dbh = $self->{dbh};
	my $id = undef;

	my $sth = $dbh->prepare(qq/
		SELECT t.id FROM 
			Network AS n, Site AS s, Sensor as t 
		WHERE s.id=? AND n.id=? AND t.name=? AND s.id = n.site_id 
		AND t.network_id = n.id/);
	$sth->execute($site_id, $network_id, lc($sensor));
	my $row = $sth->fetchrow_arrayref;
	$sth->finish;
	if (!$row) {
		$sth = $dbh->prepare(qq/INSERT OR $_replace INTO Sensor (id, name, network_id, location, a, b)
			VALUES (NULL, ?, ?, ST_GeomFromText(?, 4326), ?, ?)
			/);
		$sth->execute(lc($sensor), $network_id, sprintf('POINT(%.7f %.7f)',$long, $lat), $a, $b);
		$id = $dbh->last_insert_id(undef, undef, undef, undef);
		$dbh->commit;
	} else {
		$id = $row->[0];
	}
	$id;
}

sub _inject_measure($$$$\%$) {
	my ($self, $site_id, $network_id, $sensor_id, $record, $count_override) = @_;

	my @id = undef;
	my $dbh = $self->{dbh};
	my $timestamp = $self->{timestamp} //= $self->_timestamp;

	my $datetime = $record->{timestamp};
	my ($value, @d, $unit, $type, $depth);
	my $sth = $dbh->prepare(qq/
			INSERT OR $_replace INTO Measure (datetime, depth_min, depth_max, count,
			       sensor_id, value, type, unit, injection_datetime) 
			VALUES (?,?,?,?,?,?,?,?,?)
			/);
	for my $key (keys %$record) {
		next if ! defined $record->{$key};
		given ($key) {
			when (/^temp_/) {
				(undef, $depth) = split(/_/, $key);
				@d = $self->_depth($depth);
				$value = $record->{$key};
				$type = 'temperature';
				$unit = 'C deg';

			}
			when (/^sm_/) {
				(undef, $depth) = split(/_/, $key);
				@d = $self->_depth($depth);
				$value = $record->{$key};
				$type = 'soil_moisture';
				$unit = 'VWC';
			}
			when (/^rainfall_/) {
				(undef, $depth) = split(/_/, $key);
				@d = $self->_depth($depth);
				$value = $record->{$key};
				$type = 'rainfall';
				$unit = 'mm';
			}
			when (/^ec/) {
				(undef, $depth) = split(/_/, $key);
				@d = $self->_depth($depth);
				$value = $record->{$key};
				$type = 'ec';
				$unit = 'mS/cm';
			}
			when (/^evapo/) {
				(undef, $depth) = split(/_/, $key);
				@d = $self->_depth($depth);
				$value = $record->{$key};
				$type = 'evapo';
				$unit = 'mm/s';
			}
			default { next }
		}
		my ($dmin, $dmax, $count) = map { defined $_ ? ($_ ? $_ : "0") : "NULL" } @d;
		$count = $count_override if defined $count_override;
		$dmax = $dmin if $dmax < $dmin;
		warn qq/$datetime, $dmin, $dmax, $count, $sensor_id, $value, $type, $unit, $timestamp/ if $Verbose;
		$sth->execute($datetime, $dmin, $dmax, $count, $sensor_id, $value, $type, $unit, $timestamp);
		push @id, $dbh->last_insert_id(undef, undef, undef, undef);
	}
	@id;
}

sub detach_file($$$$) {
	my ($self, $path, $target, $uuid) = @_;
	my $stripped = $path;
	my $hash = substr($uuid, 0, 4);
	my $subdir = File::Spec->catfile($target, $hash);
	mkdir $subdir or Carp::croak qq/Cannot create dir $subdir: $!/ if ! -d  $subdir;
	$stripped =~ s/(.*)\.[^.]+$/$1/;
	if ($Link) {
		my $val = link $path, File::Spec->catfile($subdir, $uuid);
		my $ret1 = link "$path.hdr", File::Spec->catfile($subdir, "$uuid.hdr") if $val && -s "$path.hdr";
		my $ret2 = link "$stripped.hdr", File::Spec->catfile($subdir, "$uuid.hdr") if $val && -s "$stripped.hdr";
		return $val;
	} else {
		my $val = copy($path, File::Spec->catfile($subdir, $uuid));
		my $ret1 = copy("$path.hdr", File::Spec->catfile($subdir, "$uuid.hdr")) if $val && -s "$path.hdr";
		my $ret2 = copy("$stripped.hdr", File::Spec->catfile($subdir, "$uuid.hdr")) if $val && -s "$stripped.hdr";
		return $val;
	}
}

sub clean_file($$$) {
	my ($self, $target, $uuid) = @_;
	my $val = unlink File::Spec->catfile($target, $uuid);
	$val = unlink File::Spec->catfile($target, "$uuid.hdr") if $val && -s File::Spec->catfile($target, "$uuid.hdr");
	$val;
}

#
# Object destructor needs to get rid of the db connection.
#

sub DESTROY($) {
	my $self = shift;
	$self->close if $self;
}

#
# From here on all public methods are defined. 
#

sub to_depth($$$$) {
	my ($self, $str, $from, $to) = @_;
	$str = $str.sprintf("_%.2fT%.2f", $from, $to) =~ s/\./d/gr;
	$str;
}

sub correct_temperature($$) {
	my ($self, $temp) = @_;

	if ($temp >= 0 and $temp <= 1022) {
		if ($temp >= 900) {
			$temp = 900 + 5*($temp - 900);
		}
		$temp = ($temp - 400) / 10.0;
	}
	$temp;
}

sub set_injection_datetime($) {
	my $self = shift;

	$self->{timestamp} = $self->_timestamp;
}


use DateTime::Format::DateParse;

sub normalize_datetime($$) {
	my ($self, $timestamp) = @_;

	my $dt = DateTime::Format::DateParse->parse_datetime($timestamp, 'UTC');
	Carp::croak qq/Wrong format in date $timestamp/ if ! $dt;
	$dt->set_year($dt->year + 100) if $dt->year < 2000;
	sprintf qq/%s %s/, $dt->ymd, $dt->hms;
}

use DateTime::Format::ISO8601;

sub adjust_datetime($$) {
	my ($self, $timestamp) = @_;
	eval {
		my $dt = DateTime::Format::ISO8601->parse_datetime( $timestamp );
		$dt->set_time_zone('UTC');
		my $t = $dt->datetime =~ s/T/ /r;
		$timestamp = $t if $t;
	};
	warn $@ if $@;
	$timestamp;
}

use DateTime;

sub adjust_datetime_for_tz($$$) {
	my ($self, $timestamp, $tz) = @_;

	my @t = split(/[- :]/, $timestamp);
	#
	# Note that due to DST not all local times could be defined.
	# So it is better specifying TZ in terms of UTC offset instead
	# of country areas.
	#
	eval {
		my $dt = DateTime->new(
			year       => $t[0],
			month      => $t[1],
			day        => $t[2],
			hour       => $t[3] % 24,
			minute     => $t[4] % 60,
			second     => $t[5] ? $t[5] % 60 : 0,
			nanosecond => 0,
			time_zone  => $tz,
		);
		$dt->set_time_zone('UTC');
		# sometimes data contains invalid time such as 
		# 24:00 instead of +1day 00:00
		$dt->add(days => 1) if $t[3] == 24;
		my $t = $dt->datetime =~ s/T/ /r;
		$timestamp = $t if $t;
	}; 
	warn $@ if $@;
	$timestamp;
}

sub inject_ground_data($$$\%$) {
	my ($self, $site, $network, $rec, $count) = @_;
	my $dbh = $self->{dbh};

	# note that order of check-in matters
	my $site_id = $self->_site_check_in($site);
	my $network_id = $self->_network_check_in($site_id, $network);
	my $sensor_id = $self->_sensor_check_in($site_id, $network_id, 
		$rec->{sensorid}, $rec->{long}, $rec->{lat}, $rec->{a}, $rec->{b});
	my @measures = $self->_inject_measure($site_id, $network_id, $sensor_id, $rec, $count);
	@measures;
}

sub inject_raster_data($$\%) {
	my ($self, $site, $class, $rec) = @_;
	my $dbh = $self->{dbh};
	my $id = undef;
	my $timestamp = $self->{timestamp} //= $self->_timestamp;

	$class = lc($class);
	my $site_id = $self->_site_check_in($site) if $site;
	if ($class =~ m/^(image|preproc|product)$/) {
		my $processing_datetime = $rec->{processing_datetime};
		my $injection_datetime = $rec->{injection_datetime};
		$injection_datetime //= $timestamp;
		my $reference_datetime = $rec->{reference_datetime};
		my $origin = $rec->{origin};
		my $product_type = $rec->{product_type};
		my $filename = $rec->{filename};
		my $uri = $rec->{uri};
		my $boundingbox   = $rec->{boundingbox};
		my $processor   = $rec->{processor};
		my $footprint = $rec->{footprint} // $rec->{boundingbox};

		my $provider_injection_datetime = $rec->{provider_injection_datetime};
		my $platform = $rec->{platform};
		my $rel_orbitno = sprintf("%d", $rec->{rel_orbitno});
		my $orbitno = sprintf("%d", $rec->{orbitno});
		my $provider_filename = $rec->{provider_filename};
		my $provider_reference = $rec->{provider_reference};
		my $provider_hash = $rec->{provider_hash};
		my $tag = lc($rec->{tag});
		my $direction = lc($rec->{direction});

		say Dumper($rec) if $Verbose;
		my $uri_id = $self->_path_check_in($uri);
		if ($class =~ m/^image$/) {
			my $sth = $dbh->prepare(qq/ 
				INSERT OR $_replace INTO Image (id, provider_injection_datetime, injection_datetime,
				reference_datetime, platform, rel_orbitno, orbitno, product_type, provider_filename,
				provider_reference, provider_hash, filename, uri_id, footprint, tag, direction) VALUES 
				(NULL, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ST_GeomFromText(?, 4326), ?, ?)
				/
			);
			$sth->execute($provider_injection_datetime, $injection_datetime, $reference_datetime, $platform, 
						  $product_type, $rel_orbitno, $orbitno, $product_type, $provider_filename,
						  $provider_reference, $provider_hash, $filename, $uri_id, $footprint, $tag, $direction );
			$id = $dbh->last_insert_id(undef, undef, undef, undef);
		} 
		elsif ($class =~ m/^preproc$/) {
			my $sth = $dbh->prepare(qq/ 
				INSERT OR $_replace INTO ProcessedImage (id, processing_datetime, injection_datetime,
				reference_datetime, processor, product_type, filename, uri_id, footprint, boundingbox, tag, rel_orbitno, direction) VALUES 
				(NULL, ?, ?, ?, ?, ?, ?, ?, ST_GeomFromText(?, 4326), ST_GeomFromText(?, 4326), ?, ?, ?)
				/
			);
			$sth->execute($processing_datetime, $injection_datetime, $reference_datetime, $processor, $product_type,
						  $filename, $uri_id, $footprint, $boundingbox, $tag, $rel_orbitno, $direction);
			$id = $dbh->last_insert_id(undef, undef, undef, undef);
		}
		else {
			my $sth = $dbh->prepare(qq/ 
				INSERT OR $_replace INTO Product (id, processing_datetime, injection_datetime, 
				reference_datetime, processor, product_type, filename, uri_id, footprint, boundingbox, tag, rel_orbitno, direction) VALUES 
				(NULL, ?, ?, ?, ?, ?, ?, ?, ST_GeomFromText(?, 4326), ST_GeomFromText(?, 4326), ?, ?, ?)
				/
			);
			$sth->execute($processing_datetime, $injection_datetime, $reference_datetime, $processor, $product_type,
						  $filename, $uri_id, $footprint, $boundingbox, $tag, $rel_orbitno, $direction);
			$id = $dbh->last_insert_id(undef, undef, undef, undef);
		}

	} else {
		Carp::carp qq/Missing or wrong raster class, use image, preproc or product/;
	}
	$id;
}

sub get_sensors($$$) {
	my ($self, $site, $network) = @_;
	my $dbh = $self->{dbh};

	my %sensors = ();
	my $sth;
	if ($site && $network) {
		$sth = $dbh->prepare(qq/
			SELECT s.name, ST_X(s.location) AS long, ST_Y(s.location) AS lat, s.a, s.b, n.name as network
			FROM Sensor AS s, Site AS t, Network AS n 
			WHERE s.network_id=n.id AND n.site_id=t.id AND t.name=? AND n.name=?/);
		$sth->execute($site, $network);
	} elsif ($network) {
		$sth = $dbh->prepare(qq/
			SELECT s.name, ST_X(s.location) AS long, ST_Y(s.location) AS lat, s.a, s.b, n.name as network
			FROM Sensor AS s, Network AS n
			WHERE s.network_id=n.id AND n.name=?/);
		$sth->execute($network);
	} elsif ($site) {
		$sth = $dbh->prepare(qq/
			SELECT s.name, ST_X(s.location) AS long, ST_Y(s.location) AS lat, s.a, s.b, n.name as network 
			FROM Sensor AS s, Site AS t, Network AS n
			WHERE s.network_id=n.id AND n.site_id=t.id AND t.name=?/);
		$sth->execute($site);
	} else {
		$sth = $dbh->prepare(qq/
			SELECT s.name, ST_X(s.location) AS long, ST_Y(s.location) AS lat, s.a, s.b, n.name as network 
			FROM Sensor AS s, Network AS n
			WHERE s.network_id=n.id/);
		$sth->execute;
	}
	while (my $row = $sth->fetchrow_arrayref) {
		$sensors{lc($row->[0])} = {
			longitude => $row->[1],
			latitude  => $row->[2],
			a => $row->[3],
			b => $row->[4],
			network => $row->[5],
		};
	}
	%sensors;
}

sub export_ground_data($%) {
	my $self = shift;
	my %args = (
		site => undef,
		network => undef,
		from => undef,
		to => undef,
		aliases => undef,
		type => undef,
		output_function => undef,
		@_
	);
	my $dbh = $self->{dbh};
	my $condition;
	my $aliases = {};
	$aliases = $args{aliases} if defined($args{aliases});
	my $outfunc = undef;
	$outfunc = $args{output_function} if defined($args{output_function});

	if (defined($args{network})) {
		my $network = $args{network};
		$condition .= sprintf(" AND m.sensor_id IN (SELECT id FROM Sensor WHERE network_id IN (SELECT id FROM NETWORK WHERE name=%s))", $dbh->quote($network));
	}
	if (defined($args{site})) {
		my $site = $args{site};
		$condition .= sprintf(" AND m.sensor_id IN (SELECT id FROM Sensor WHERE network_id IN (SELECT id FROM Network WHERE site_id in (SELECT id from Site WHERE name=%s)))", $dbh->quote($site));
	}
	if (defined($args{type})) {
		my $type = $args{type};
		$condition .= sprintf(" AND m.type=%s",$dbh->quote($type));
	}

	my $min_date = undef;
	my $max_date = undef;
	if (defined($args{from})) {
		my $from = $self->normalize_datetime($args{from});
		$condition .= sprintf(" AND date(m.datetime) >= %s",$dbh->quote($from));
		$min_date = (split(/ /, $from))[0];
	} else {
		my $query_mindate = qq/
		SELECT DATE(MIN(m.datetime)) 
		FROM Measure AS m, Sensor as s, Network as n, Site as t 
		WHERE m.sensor_id=s.id AND s.network_id=n.id AND n.site_id=t.id $condition/;
		my $sth = $dbh->prepare($query_mindate);
		$sth->execute;
		$min_date = $sth->fetchrow_arrayref->[0];
	}
	if (defined($args{to})) {
		my $to = $self->normalize_datetime($args{to});
		$condition .= sprintf(" AND date(m.datetime) <= %s",$dbh->quote($to));
		$max_date = (split(/ /, $to))[0];
	} else {
		my $query_maxdate = qq/
		SELECT DATE(MAX(m.datetime)) 
		FROM Measure AS m, Sensor as s, Network as n, Site as t 
		WHERE m.sensor_id=s.id AND s.network_id=n.id AND n.site_id=t.id $condition/;
		my $sth = $dbh->prepare($query_maxdate);
		$sth->execute;
		$max_date = $sth->fetchrow_arrayref->[0];
	}

	my $query = qq/
		SELECT 	m.*, s.name AS station, s.a AS a, s.b AS b, m.value*s.a+s.b as calibrated, ST_X(s.location) as x, ST_Y(s.location) as y, s.network_id AS network_id, n.site_id AS site_id, 
				n.name AS network_name, t.name AS site_name 
		FROM Measure AS m, Sensor as s, Network as n, Site as t 
		WHERE m.sensor_id=s.id AND s.network_id=n.id AND n.site_id=t.id $condition ORDER BY s.name, m.datetime/;
	my $sth = $dbh->prepare($query);
	$sth->execute;
	while (my $hash = $sth->fetchrow_hashref) {
		$hash->{calibrated} //= $hash->{value};
		next if $hash->{calibrated} < 0; 
		my $station = $hash->{station};
		$hash->{station} = uc($aliases->{$station}) if defined($aliases->{$station});
		print Dumper($hash) if $Verbose;
		&$outfunc(
			data => $hash,
			from => $args{from},
			to => $args{to},
			min_date => $min_date,
			max_date => $max_date,
		) if defined($outfunc);
	}
	($min_date, $max_date);
}

sub get_networks($$) {
	my $self = shift;
	my $site = shift;
	my $dbh = $self->{dbh};

	my @networks = ();
	my $sth;
	if ($site) {
		$sth = $dbh->prepare(qq/SELECT n.name FROM Network AS n, Site AS s 
			WHERE n.site_id=s.id AND s.name=?/);
		$sth->execute($site);
	} else {
		$sth = $dbh->prepare(qq/SELECT name FROM Network/);
		$sth->execute;
	}
	while (my $row = $sth->fetchrow_arrayref) {
		push @networks, $row->[0];
	}
	@networks;
}

sub get_sites($) {
	my $self = shift;
	my $dbh = $self->{dbh};

	my @sites = ();
	my $sth = $dbh->prepare(qq/SELECT name FROM Site/);
	$sth->execute;
	while (my $row = $sth->fetchrow_arrayref) {
		push @sites, $row->[0];
	}
	@sites;
}

sub close($) {
	my $self = shift;
	my $dbh = $self->{dbh};
	$dbh->disconnect if $dbh;
}

sub commit($) {
	my $self = shift;
	my $dbh = $self->{dbh};
	$dbh->commit if $dbh;
}

sub rollback($) {
	my $self = shift;
	my $dbh = $self->{dbh};
	$dbh->rollback if $dbh;
}

sub db_handle($) {
	my $self = shift;
	$self->{dbh};
}

sub init_repo_path($$$$$) {
	my ($self, $ini_repo_id, $ini_repo_path, $repository_id, $repository_path ) = @_;
	my $dbh = $self->{dbh};
	my ($id, $path) = (undef, undef);

	if (defined($repository_id) || defined($repository_path)) {
		if (defined($repository_id)) {
			my $sth = $dbh->prepare(qq/SELECT id, uri FROM Path WHERE id=?/);
			$sth->execute($repository_id);
			my $row = $sth->fetchrow_hashref;
			if ($row && $row->{uri}) {
				$id = $row->{id};
				$path = $row->{uri};
			}
		}
		if (defined($repository_path)) {
			$id = $self->_path_check_in($repository_path);
			$path = $repository_path;
		}
	} else {
			if (defined($ini_repo_id) || defined($ini_repo_path)) {
				if (defined($ini_repo_id)) {
					my $sth = $dbh->prepare(qq/SELECT id, uri FROM Path WHERE id=?/);
					$sth->execute($ini_repo_id);
					my $row = $sth->fetchrow_hashref;
					if ($row && $row->{uri}) {
						$id = $row->{id};
						$path = $row->{uri};
					}
				}
				if (defined($ini_repo_path)) {
					$id = $self->_path_check_in($ini_repo_path);
					$path = $ini_repo_path;
				}
			}
	}
	make_path($path) if $path;
	($id, $path);
}

END { }       # module clean-up code here (global destructor)

1;

__END__

=head1 NAME

SSM_Tools::Database -- a module to manage SSM-DB implementation details

=head1 SYNOPSIS

  use SSM_Tools::Database;

  $db = SSM_Tools::Database->new($filename);

  $db->inject_ground_data($site, $network, $datarec);

  $db->close();

  $db->commit();

  $db->rollback();

  $db->adjust_datetime_for_tz($timestamp, $tz);

  $db->set_injection_datetime();

  $db->db_handle();

  $db->get_sites();

  $db->get_networks($site);

  $db->get_sensors($site, $network);

  $SSM_Tools::Database::Verbose = 0* | 1
  $SSM_Tools::Database::Replace = 0* | 1

=head1 DESCRIPTION

This module provides all object-oriented methods required to inject,
query and extract satellite and ground information into the SSM DB.
All details related to the DB implementation are hidden in this module
so that other modules should not use any SQL query or depend on the
specific engine used. Current implementation is based on Spatialite and
depends on the DBI interface provided for SQLite.

=head2 Class Variables and Methods

A list with descriptions of the module public methods follows. Note
that the package provides only a full OO interface, no exported subs.

=head3 SSM_Tools::Database->new($filename)

Returns a database object bound to a defined Spatialite $filename. It
creates the database if it does not exist or connect to an existing one.
In case of a new database only Spatialite meta tables are created. The
SSM scheme needs to be created after that, so it is better creating the
database in advance with the provided C<build.sql> script. That could
be done with a shell command like:

  rm -f $filename && spatialite -init build.sql $filename .quit

=head3 $db->inject_ground_data($site, $network, $datarec)

Inject a ground data entry into the SSM database. You need to specify a
site and a network (possibly already existent). The C<$datarec> hash ref
follows a predefined scheme that needs to be respected. E.g.:

	timestamp               => '2016-04-01 14:59:00',
	sensorid                => '<id>:<alias>',
	long                    => '16.4484848',
	lat                     => '41.3949494',
	temp_<depth_range>      => '21.2:0',
	sm_<depth_range>        => '323',
	rainfall_<depth_range>  => '12',
	ec                      => 0.49494
	a                       => 12.303
	b                       => 494.000

Multiple temperature and SSM value at different depth or depth range can
be specified in compact way. The SSM value can be calibrated or not, on
the basis of the pair of C<a> and C<b> coefficients.

The general form of a depth range is something like: 2d34T5d4:1 which
means a range like [2.34..5.4]. Decimal parts and the right limit can
are optional. So it is also acceptable specifying 2d34 or 2T5 or a
simple 2. The (optional) successive counter (e.g. :1 in the previous
example) is used when multiple measures are taken within the same depth
range, so that they are counted consecutively to distinguish all of
them.

The BNF form for the depth range syntax is the following:

  depth_range ::= <depth_value> [ : <count> ] | 
                  <depth_value> T <depth_value> [ : <count> ] | : <count>
  depth_value ::= <digit>+ | <digit>+ d <digit>+
  count       ::= <digit>+

=head3 $db->inject_raster_data($site, $class, $datarec);

This general method is used to inject various classes of raster data, among
raw images, pre-processed/itermediate images and final products. The $class
argument can assume 'image', 'preproc' or 'product' values to distinguish
among the different kind of raster data to be considered. Most of the
$datarec attributes are the same for any kind of raster objects, but some
attributes are per-object specific.

=head3 $db->set_injection_datetime();

This method set at B<now> the injection timestamp. This is done automagically
at the first injection of data, but the timestamp can be reset on demand
for any specific purpose. The same timestamp will be used for all successive
injections of new data.

=head3 $db->close()

=head3 $db->commit()

=head3 $db->rollback()

All these methods refer to corresponding DBI methods to store/discard changes
into the database or disconnect. Note that transactions are automatically
activated to improve performances, but that implies that explicit commtting
has to be performed before C<close()>.

=head3 $db->adjust_datetime_for_tz($timestamp, $tz);

Some sensors use local time instead of UTC/GMT time for each measure.
This utility function is used to normalize the time to UTC and 
should be used before storing any timestamps in the db. The $tz string
variable is a standard timezone specification, e.g. C<Europe/Rome>.
The $timestamp variable is an ISO datetime string specification, i.e.
one of the following forms:

  YYYY-MM-DD HH:MM:SS
  YYYY-MM-DD HH:MM

So, the timestamp specified in C<inject_ground_data()> should be already
normalized to UTC time zone. Note that the timezone used can include
or not a change for DST too, automatically managed. The use of DST 
could potentially give problems if invalid times are used in measures:
some times could not exist or be repeated twice. Currently, these 
events cause an exception that is captured but ignored, and the timestamp
is not changed. See C<DateTime> module for more information.

=head3 $db->db_handle();

This accessor function returns the DBI handle for any other possible
use. Users should avoid to use this handle, it is provided only for 
trickery, but trickery should be avoided, as principle.

=head3 $db->get_sites();

=head3 $db->get_networks($site);

=head3 $db->get_sensors($site, $network);

This methods list all unique names for sites, networks and sensors.
Args can be C<undef> and the resulting unique labels will ignore
the site and/or network considered. This could make sense or not.

=head3 $Verbose = 0* | 1;

This variable allow to change the module level of verbosity.
The default (0) make it quiet, but for exceptions. This variable
can be explicity exported.

=head3 $Replace = 0* | 1;

This variable allows to REPLACE on INSERT an entry. The default (0)
implies an IGNORE. This is tipically only considered for measures.

=head3 $Link = 0 | 1*;

This variable allows linking on detach for files. The default (1)
implies that files and repository belong to the same device and the
device allows hard linking by inode. That is not possible with
NFS shares, for instance. The only alternative is copying into 
the repository which could result in a wasting of disk space.

=head1 AUTHOR

Francesco P. Lovergine

=head1 COPYRIGHT

Copyright (C) 2017-2021 Francesco P. Lovergine <francesco.lovergine@cnr.it>
 
=cut
