#!/bin/bash
#
# This script extracts statistics on a site by grouping together sensors
# within the same cell of a certain size (1km or 3km).
# 
# This script can also be called with two optional argument that is the site
# name and the 'tag' of a SM injection.
# 

usage() {
	echo "usage: $0 <site> <label> [asc|desc|ascdesc] [sm|s0]"
	exit 1
}

L=run_n4_v5_5_20180105_pix520m_flag12
S=yanco
O=
P=Product

if [ $# -eq 1 ]; then
	S="$1"
fi
if [ $# -eq 2 ]; then
	S="$1"
	L="$2"
fi
if [ $# -eq 3 ]; then
	S="$1"
	L="$2"
	O="$3"
	case $O in
		asc|desc|ascdesc)
			;;
		*)
			usage
			;;
	esac
fi
if [ $# -eq 4 ]; then
	S="$1"
	L="$2"
	O="$3"
	case $O in
		asc|desc|ascdesc)
			;;
		*)
			usage
			;;
	esac
	P="$4"
	case $P in
		sm)
			P=Product
			;;
		s0)
			P=ProcessedImage
			;;
		*)
			usage
			;;
	esac
fi

$SM_TOOLS/etc/grass_check.sh || exit 1

cat <<EOF
Site: $S
Tag: $L
Orientation: $O
EOF

set -e

D=$SM_TOOLS_OUTPUTDIR
T=$SM_TOOLS_TMPDIR
R=$SM_TOOLS_REPO
N=$(date +%Y%m%d)

# Prepare WKT geometries 
source $SM_TOOLS/misc/gen_${S}_grid_polys.sh $O
   
# For each group aka grid cell extract average measures
for i in $(seq 1 $NCATS) 
do
echo Group $i: ${P[i]}
s=$(echo ${S[i]}|sed -e 's/ /,/g')
sql=`cat <<EOF | sed -e "s/@POLYGON@/${U[i]}/" -e "s/@SENSOR_IDS@/${s}/" -e "s/@TAG@/${L}/"
select p.filename as FileName,
	   date(p.reference_datetime) as RefDate,
	   strftime('%H:%M', p.reference_datetime) as RefTime,
	   avg(s.a*m.value+s.b) as avg,
	   stddev_pop(s.a*m.value+s.b) as stddev,
	   count(m.value) as ct,
	   min(m.depth_max) as Depth,
	   date(min(m.datetime)) as minMeasureDate,
	   strftime('%H:%M', min(m.datetime)) as minMeasureTime,
	   date(max(m.datetime)) as maxMeasureDate,
	   strftime('%H:%M', max(m.datetime)) as maxMeasureTime
from $P as p,
	 Measure as m,
	 Sensor as s
where m.count=0 and m.type = 'soil_moisture' and s.id = m.sensor_id and
	  date(m.datetime) = date(p.reference_datetime) and 
	  datetime(m.datetime) >= datetime(p.reference_datetime,'-1 hour') and 
	  datetime(m.datetime) <= datetime(p.reference_datetime) and 
      m.depth_max <= 0.05 and 
	  ST_Contains(p.boundingbox, s.location) and 
	  p.tag = '@TAG@' and
	  m.sensor_id in (@SENSOR_IDS@)
group by 1,2,3
order by RefDate, Reftime; 
EOF
`
echo $sql | spatialite -noheader $SM_TOOLS_DB >${T}/${N}_${S}_raw_stats_cell_G${i}_${O}.txt
done

# Now create statistics for every grid cell 
for i in $(seq 1 $NCATS) 
do
	if [ $($G g.list t=vect|grep -E "^${S}_G${i}"|wc -l) != 1 ];
	then
		# Create vector geometries starting from WKTs
		np=$(echo ${G[i]}|sed -e 's/POLYGON/ /' -e 's/((//' -e 's/))//' -e 's/,/\n/g'|wc -l)
		cat >$T/$$.txt <<EOF && echo ${G[i]}|sed -e 's/POLYGON/ /' -e 's/((//' -e 's/))//' -e 's/,/\n/g' >>$T/$$.txt
B $np
EOF
		$G v.in.ascii -n --o input=$T/$$.txt output=${S}_G${i} format=standard --q
		rm -f $T/$$.txt
	fi
	# Append image statistics
	echo "Group|S-1 Date|S-1 Time|Mean Calibrated SM|Pop-StdDev Calibrated SM|Count|Max Depth(m)|Min Station Date|Time|Max Station Date|Time|SMOSAR Mean|SMOSAR StdDev|SMOSAR Min|SMOSAR Max|Cells|NULL cells" >$D/${N}_${S}_stats_bycell_G${i}_${O}.txt
	for line in $(cat ${T}/${N}_${S}_raw_stats_cell_G${i}_${O}.txt)
	do
		file=$(echo $line|cut -d\| -f1)
		rest=$(echo $line|cut -d\| -f2-)
		[ $($G g.list t=rast pat="$file"|wc -l) -eq 1 ] || $G r.external --o --quiet input=$R/$file output=$file
		$G g.region --quiet rast=$file 
		$G g.region --quiet vect=${S}_G${i} -a
		eval $($G r.univar map=$file -g)
		echo "G${i}|$rest|$mean|$stddev|$min|$max|$cells|$null_cells" >>$D/${N}_${S}_stats_bycell_G${i}_${O}.txt
	done
done 

$G_CLEAN

exit 0
