-- Copyright (C) 2017-2022 Francesco P. Lovergine <francesco.lovergine@cnr.it>
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License version 2, 
-- as published by the Free Software Foundation.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
--

-- !!!! Run under the Spatialite tool, not the SQLite one.

--
-- Initialize an empty Spatialite database for SSMDB applications, including
-- all geospatial columns. This is tested with Spatialite 4.1+
--

BEGIN TRANSACTION;

--
-- Table Path
--
DROP TABLE IF EXISTS `Path`;
CREATE TABLE `Path` (
	`id` INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT, 
    `uri` TEXT NOT NULL
);
CREATE UNIQUE INDEX `idx_path_name` ON `Path`(`uri`);

--
-- Table Site
--
SELECT DisableSpatialIndex('Site', 'aoi');
SELECT DiscardGeometryColumn('Site','aoi');
DROP TABLE IF EXISTS `Site`;
CREATE TABLE `Site` (
	`id` INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT, 
    `name` TEXT NOT NULL,
	`priority` INTEGER NOT NULL DEFAULT 0
);
CREATE UNIQUE INDEX `idx_site_name` ON `Site`(`name`);
CREATE INDEX `idx_site_prio` ON `Site`(`priority`);
SELECT AddGeometryColumn( 'Site', 'aoi', 4326, 'MULTIPOLYGON', 'XY');
SELECT CreateSpatialIndex('Site', 'aoi');

/*
 *
 *	These tables are used for implementation of ground data support
 *
 */

--
-- Table Network
--
DROP TABLE IF EXISTS `Network`;
CREATE TABLE `Network` (
    `id` INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT, 
    `name` TEXT NOT NULL,
	`site_id` UNSIGNED INT NOT NULL,
  	CONSTRAINT `fk_network_site`
		FOREIGN KEY(`site_id`) REFERENCES `Site`(`id`)
);
CREATE UNIQUE INDEX `idx_network` ON `Network` (`name`);
CREATE INDEX `idx_site` ON `Network`(`site_id`);

--
-- Table Sensor
--
SELECT DisableSpatialIndex('Sensor', 'location');
SELECT DiscardGeometryColumn('Sensor','location');
DROP TABLE IF EXISTS `Sensor` ;
CREATE TABLE IF NOT EXISTS `Sensor` (
  `id` INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT,
  `name` TEXT NOT NULL,
  `network_id` INT UNSIGNED NOT NULL,
  `a` REAL,
  `b` REAL,
  CONSTRAINT `fk_sensor_network`
  	FOREIGN KEY (`network_id`) REFERENCES `Network`(`id`)
);
CREATE UNIQUE INDEX `idx_sensor_name` ON `Sensor`(`name`,`network_id`);
CREATE INDEX `idx_fk_network` ON `Network` (`id`);
SELECT AddGeometryColumn( 'Sensor', 'location', 4326, 'POINT', 'XY');
SELECT AddGeometryColumn( 'Sensor', 'polygon', 4326, 'POLYGON', 'XY');
SELECT CreateSpatialIndex('Sensor', 'location');
SELECT CreateSpatialIndex('Sensor', 'polygon');

--
-- Table SensorParam
--
DROP TABLE IF EXISTS `SensorParam` ;
CREATE TABLE IF NOT EXISTS `SensorParam` (
  `sensor_id` INT UNSIGNED NOT NULL,
  `type` TEXT NOT NULL,
  `depth_min` REAL,
  `depth_max` REAL,
  `a` REAL,
  `b` REAL,
  CONSTRAINT `fk_sensor_param`
  	FOREIGN KEY (`sensor_id`) REFERENCES `Sensor`(`id`),
  CONSTRAINT d CHECK(depth_min <= depth_max)
);
CREATE INDEX `idx_sp_sensor_id` ON `SensorParam` (`sensor_id`);
CREATE INDEX `idx_sp_dmin` ON `SensorParam` (`depth_min`);
CREATE INDEX `idx_sp_dmax` ON `SensorParam` (`depth_max`);
CREATE INDEX `idx_sp_type` ON `SensorParam` (`type`);
CREATE UNIQUE INDEX `idx_sensor_param` ON `SensorParam` (
	`sensor_id`, `type`, `depth_min`, `depth_max`
);

--
-- Table Measure
--
DROP TABLE IF EXISTS `Measure` ;
CREATE TABLE IF NOT EXISTS `Measure` (
  `datetime` TEXT NOT NULL,
  `depth_min` REAL,
  `depth_max` REAL,
  `sensor_id` INT UNSIGNED NOT NULL,
  `value` REAL,
  `count` INT UNSIGNED DEFAULT NULL,
  `type` TEXT NOT NULL,
  `unit` TEXT NOT NULL,
  `injection_datetime` TEXT NOT NULL,
  CONSTRAINT `fk_measure_sensor`
  	FOREIGN KEY (`sensor_id`) REFERENCES `Sensor`(`id`),
  CONSTRAINT c CHECK(depth_min <= depth_max)
);
CREATE INDEX `idx_type` ON `Measure` (`type`);
CREATE INDEX `idx_sensor_id` ON `Measure` (`sensor_id`);
CREATE INDEX `idx_datetime` ON `Measure` (`datetime`);
CREATE UNIQUE INDEX `idx_measure` ON `Measure` (
	`datetime`, `type`, `sensor_id`, `depth_min`, `depth_max`, `count`
);
CREATE INDEX `idx_injection` ON `Measure` (`injection_datetime`);

/*
 *
 *	These tables are used for implementation of satellite data and derived
 *	maps support
 *
 */

--
-- Table Collection
--
DROP TABLE IF EXISTS `Collection`;
CREATE TABLE IF NOT EXISTS `Collection` (
	`id` INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT,
	`description` TEXT NOT NULL
);
CREATE INDEX `idx_collection_desc` ON `Collection` (`description`);

--
-- Relation CollectionSite
--
DROP TABLE IF EXISTS `CollectionSite`;
CREATE TABLE IF NOT EXISTS `CollectionSite` (
	`collection_id` INTEGER NOT NULL,
	`site_id` INTEGER NOT NULL,
	CONSTRAINT fk_collection_id
		FOREIGN KEY (`collection_id`) REFERENCES `Collection`(`id`),
  	CONSTRAINT `fk_site_id`
		FOREIGN KEY(`site_id`) REFERENCES `Site`(`id`)
);
CREATE UNIQUE INDEX `idx_collection_site` 
	ON `CollectionSite` (`collection_id`,`site_id`);

--
-- Relation CollectionImage
--
DROP TABLE IF EXISTS `CollectionImage`;
CREATE TABLE IF NOT EXISTS `CollectionImage` (
	`collection_id` INTEGER NOT NULL,
	`image_id` INTEGER NOT NULL,
	CONSTRAINT fk_coll_id
		FOREIGN KEY (`collection_id`) REFERENCES `Collection`(`id`),
  	CONSTRAINT `fk_img_id`
		FOREIGN KEY(`image_id`) REFERENCES `Image`(`id`)
);
CREATE UNIQUE INDEX `idx_collection_image` 
	ON `CollectionImage` (`collection_id`,`image_id`);

--
-- Table Image
--
DROP TABLE IF EXISTS `Image`;
CREATE TABLE IF NOT EXISTS `Image` (
	`id` INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT,
	`provider_injection_datetime` TEXT NOT NULL,
	`injection_datetime` TEXT NOT NULL,
	`reference_datetime` TEXT NOT NULL,
	`platform` TEXT NOT NULL,
	`rel_orbitno` TEXT NOT NULL,
	`orbitno` NOT NULL,
	`product_type` TEXT NOT NULL,
	`provider_filename` TEXT NOT NULL,
	`filename` TEXT NOT NULL,
	`provider_reference` TEXT NOT NULL,
	`provider_hash` TEXT NOT NULL,
	`uri_id` INTEGER NOT NULL,
	`direction` TEXT NOT NULL,
	CONSTRAINT fk_image_path
		FOREIGN KEY (`uri_id`) REFERENCES `Path`(`id`)
);
CREATE INDEX `idx_inj_timestamp` ON `Image` (`injection_datetime`);
CREATE INDEX `idx_product_type` ON `Image` (`product_type`);
CREATE INDEX `idx_platform` ON `Image` (`platform`);
CREATE INDEX `idx_ron` ON `Image` (`rel_orbitno`);
CREATE INDEX `idx_provider_ref` ON `Image` (`provider_reference`);
CREATE INDEX `idx_ref_timestamp` ON `Image` (`reference_datetime`);
CREATE INDEX `idx_dir` ON `Image` (`direction`);
CREATE UNIQUE INDEX `idx_provider_fname` ON `Image` (`provider_filename`);
SELECT AddGeometryColumn( 'Image', 'footprint', 4326, 'POLYGON', 'XY');
SELECT CreateSpatialIndex('Image', 'footprint');

--
-- Table Stack
--
DROP TABLE IF EXISTS `Stack`;
CREATE TABLE IF NOT EXISTS `Stack` (
	`id` INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT,
	`description` TEXT NOT NULL,
	`master_reference` INTEGER NOT NULL,
	`collection_id` INTEGER NOT NULL,
	CONSTRAINT fk_stack_image
		FOREIGN KEY (`master_reference`) REFERENCES `Image`(`id`),
	CONSTRAINT fk_stack_collection
		FOREIGN KEY (`collection_id`) REFERENCES `Collection`(`id`)
);

--
-- Relation ImageStack
--
DROP TABLE IF EXISTS `ImageStack`;
CREATE TABLE IF NOT EXISTS `ImageStack` (
	`stack_id` INTEGER NOT NULL,
	`image_id` INTEGER NOT NULL,
	CONSTRAINT fk_stack_id
		FOREIGN KEY (`stack_id`) REFERENCES `Stack`(`id`),
	CONSTRAINT fk_image_id
		FOREIGN KEY (`image_id`) REFERENCES `Image`(`id`)
);
CREATE UNIQUE INDEX `idx_image_stack` 
	ON `ImageStack` (`stack_id`,`image_id`);

--
-- Table ProcessedImage
-- 
DROP TABLE IF EXISTS `ProcessedImage`;
CREATE TABLE IF NOT EXISTS `ProcessedImage` (
	`id` INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT,
	`processing_datetime` TEXT NOT NULL,
	`injection_datetime` TEXT NOT NULL,
	`reference_datetime` TEXT NOT NULL,
	`processor` TEXT NOT NULL,
	`product_type` TEXT NOT NULL,
	`filename` TEXT NOT NULL,
	`uri_id` INTEGER NOT NULL,
	`rel_orbitno` TEXT NOT NULL,
	`direction` TEXT NOT NULL,
	CONSTRAINT fk_pimage_path
		FOREIGN KEY (`uri_id`) REFERENCES `Path`(`id`)
);
CREATE INDEX `idx_ronpi` ON `ProcessedImage` (`rel_orbitno`);
CREATE INDEX `idx_dirpi` ON `ProcessedImage` (`direction`);
CREATE UNIQUE INDEX `idx_procimage` ON `ProcessedImage` (
	`processing_datetime`, `product_type`, `reference_datetime`, `processor`
);
SELECT AddGeometryColumn( 'ProcessedImage', 'footprint', 4326, 'POLYGON', 'XY');
SELECT CreateSpatialIndex('ProcessedImage', 'footprint');
SELECT AddGeometryColumn( 'ProcessedImage', 'boundingbox', 4326, 'POLYGON', 'XY');
SELECT CreateSpatialIndex('ProcessedImage', 'boundingbox');

--
-- Table Product
-- 
DROP TABLE IF EXISTS `Product`;
CREATE TABLE IF NOT EXISTS `Product` (
	`id` INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT,
	`processing_datetime` TEXT NOT NULL,
	`injection_datetime` TEXT NOT NULL,
	`reference_datetime` TEXT NOT NULL,
	`product_type` TEXT NOT NULL,
	`processor` TEXT NOT NULL,
	`filename` TEXT NOT NULL,
	`uri_id` INTEGER NOT NULL,
	`rel_orbitno` TEXT NOT NULL,
	`direction` TEXT NOT NULL,
	CONSTRAINT fk_product_path
		FOREIGN KEY (`uri_id`) REFERENCES `Path`(`id`)
);
CREATE INDEX `idx_ronp` ON `Product` (`rel_orbitno`);
CREATE INDEX `idx_dirp` ON `Product` (`direction`);
CREATE UNIQUE INDEX `idx_prod` ON `Product` (
	`processing_datetime`, `product_type`, `reference_datetime`, `processor`
);
SELECT AddGeometryColumn( 'Product', 'footprint', 4326, 'POLYGON', 'XY');
SELECT CreateSpatialIndex('Product', 'footprint');
SELECT AddGeometryColumn( 'Product', 'boundingbox', 4326, 'POLYGON', 'XY');
SELECT CreateSpatialIndex('Product', 'boundingbox');

--
-- Relation ProcessedImageProduct
--
DROP TABLE IF EXISTS `ProcessedImageProduct`;
CREATE TABLE IF NOT EXISTS `ProcessedImageProduct` (
	`product_id` INTEGER NOT NULL,
	`processed_image_id` INTEGER NOT NULL,
	CONSTRAINT fk_product_id
		FOREIGN KEY (`product_id`) REFERENCES `Product`(`id`),
	CONSTRAINT fk_pimage_id
		FOREIGN KEY (`processed_image_id`) REFERENCES `ProcessedImage`(`id`)
);
CREATE UNIQUE INDEX `idx_processed_image_product` 
	ON `ProcessedImageProduct` (`product_id`,`processed_image_id`);


COMMIT;
