#!/bin/sh
#
# This scripts extracts zonal statistics for LAI maps in Apulian Tavoliere
# using base land cover maps for the area which is available in different months.
#


#	 SENSAGRI LAI green & LAI brown product. GPR retrieval algorithm version 0.30. Software version 0.20.
#	 Processing Date: 12-Nov-2018. Processing Time [s]: 173
#	 Band 1 = LAI Total [m2/m2] including uncertainty thresholds masks of LAI green (40) and LAI brown (40). Masked pixels are 'NaN'.
#	 Band 2 = LAI green [m2/m2].
#	 Band 3 = LAI green Standard Deviation (SD) [m2/m2].
#	 Band 4 = LAI green Coefficient of Variation (CV) used as uncertainty threshold [%].
#	 Band 5 = LAI brown [m2/m2].
#	 Band 6 = LAI brown Standard Deviation (SD) [m2/m2].
#	 Band 7 = LAI brown Coefficient of Variation (CV) used as uncertainty threshold [%].
#	 Band 8 = Classification band: 0=Out of uncertainty threshold, 1=LAI green, 2= LAI green&brown 3=LAI brown.
#	 Band 9 = S2 SCL (Sentinel-2 Scene classification)
#	 The S2 SCL cloud mask is applied.
# 
#

set -e

TAG=lai_10m_initial_release
LC=spain
OUTDIR=$SM_TOOLS/streams_crossing/results
YEAR=2017
LAI_BANDS=1
#LAI_BANDS=$(seq 9)

# The LAI maps are 20x20m^2 wide.
MINCELLS=100

ROWS=$(spatialite -batch -noheader $SM_TOOLS_DB <<EOF
select p.filename,date(p.reference_datetime) from Product as p, Site as s 
where p.tag='$TAG' and strftime('%Y',p.reference_datetime)='$YEAR' 
and s.id=5 and ST_Intersects(p.footprint,s.aoi)>0;
EOF
)

R_CAT=$(mktemp -q)
R_CNT=$(mktemp -q)
R_MIN=$(mktemp -q)
R_MAX=$(mktemp -q)
R_AVG=$(mktemp -q)
R_STD=$(mktemp -q)
R_SKW=$(mktemp -q)
R_KUR=$(mktemp -q)

# select only some classes
$G r.reclass in=${LC}_$YEAR rules=- out=${LC}_${YEAR}_reclass --o <<EOF
1 = 1 TRIGO
4 = 4 MAIZ
70 = 70	TRIGO REGADIO
EOF

for r in $ROWS
do
	f=$(echo $r|cut -d\| -f1)
	d=$(echo $r|cut -d\| -f2)
	for i in $LAI_BANDS
	do
		echo "LAI $d $i" 
		if [ -f $OUTDIR/LAI_${d}_${i}_${LC}_stats.txt ] 
		then 
			continue
		fi

		LAI=$SM_TOOLS_REPO/$f
		$G r.in.gdal band=$i in=$LAI out=LAI --o
		$G g.region rast=LAI 
		$G r.null setnull=0 map=LAI

		# This is a way to trim a looooot of polygons  
		$G r.clump in=${LC}_${YEAR}_reclass out=${LC}_${YEAR}_clump --o
		$G r.area in=${LC}_${YEAR}_clump out=${LC}_${YEAR}_clump_1ha lesser=$MINCELLS --o
		$G r.clump in=${LC}_${YEAR}_clump_1ha out=${LC}_lc_clump --o
		$G r.mapcalc expr="${LC}_lc=${LC}_${YEAR}" --o

		# This is a trick to have a map cat -> land cover class
		#$G r.statistics base=${LC}_lc_clump cover=${LC}_lc method=mode out=LC_cat --o

		$G r.stats.zonal2 -r base=${LC}_lc_clump cover=${LC}_lc rules=$R_CAT method=min out=LAI_cat --o
		$G r.stats.zonal2 -r base=${LC}_lc_clump cover=LAI rules=$R_CNT method=count out=LAI_count --o
		$G r.stats.zonal2 -r base=${LC}_lc_clump cover=LAI rules=$R_MIN method=min out=LAI_min --o
		$G r.stats.zonal2 -r base=${LC}_lc_clump cover=LAI rules=$R_MAX method=max out=LAI_max --o
		$G r.stats.zonal2 -r base=${LC}_lc_clump cover=LAI rules=$R_AVG method=average out=LAI_mean --o
		$G r.stats.zonal2 -r base=${LC}_lc_clump cover=LAI rules=$R_STD method=stddev2 out=LAI_stddev --o
		$G r.stats.zonal2 -r base=${LC}_lc_clump cover=LAI rules=$R_SKW method=skewness2 out=LAI_skewness --o
		$G r.stats.zonal2 -r base=${LC}_lc_clump cover=LAI rules=$R_KUR method=kurtosis2 out=LAI_kurtosis --o
		echo "cat|class|crop|count|min|max|avg|stddev|skew|kurt" >$OUTDIR/LAI_${d}_${i}_${LC}_stats.txt
		for cat in $(cut -d= -f1 $R_CAT) 
		do
			class=$(grep -E "^$cat =" $R_CAT|cut -d' ' -f4|cut -d. -f1)
			count=$(grep -E "^$cat =" $R_CNT|cut -d' ' -f4|cut -d. -f1)
			# skip when number of cells is too low
			if [ $count -ge $MINCELLS ]
			then
				crop=$($G r.category --v map=${LC}_lc cat=$class sep=\||cut -d\| -f2)
				min=$(grep -E "^$cat =" $R_MIN|cut -d' ' -f4|sed -e 's/-nan/NA/')
				max=$(grep -E "^$cat =" $R_MAX|cut -d' ' -f4|sed -e 's/-nan/NA/')
				avg=$(grep -E "^$cat =" $R_AVG|cut -d' ' -f4|sed -e 's/-nan/NA/')
				stddev=$(grep -E "^$cat =" $R_STD|cut -d' ' -f4|sed -e 's/-nan/NA/')
				skew=$(grep -E "^$cat =" $R_SKW|cut -d' ' -f4|sed -e 's/-nan/NA/')
				kurt=$(grep -E "^$cat =" $R_KUR|cut -d' ' -f4|sed -e 's/-nan/NA/')
				echo "$cat|$class|$crop|$count|$min|$max|$avg|$stddev|$skew|$kurt" >>$OUTDIR/LAI_${d}_${i}_${LC}_stats.txt
			fi
		done
	done
done

rm -f $R_CAT $R_CNT $R_MIN $R_MAX $R_AVG $R_STD $R_SKW $R_KUR


$G_CLEAN
